;DIY-UHF Digital encoded signal RX
;David Lancaster
;All rights reserved

;CONFIGURABLE OPTIONS

#define ERROR_TIMEOUT			d'600000'	;approximate number of uS we'll fly on a dropped signal before shutting down
												

;END CONFIGURABLE OPTIONS


	MESSG "Assembling for 16F630"
	LIST	P=PIC16F630
	INCLUDE "P16F630.INC"
	__CONFIG _MCLRE_OFF & _CP_OFF & _WDT_OFF & _INTRC_OSC_NOCLKOUT & _CPD_OFF & _PWRTE_ON & _BODEN
	;ERRORLEVEL -302					;turn off register not in bank0 warnings


#INCLUDE ../data_format.inc


#define	SIGNAL_USE_THRESHOLD	ERROR_TIMEOUT/((PFM_FREQ)*d'255')
#define	PFM_FREQ				d'256'-d'50'		;min limit on PFM rate
#define PFM_ON_TIME				d'10'

#define NEW_INPUTS		MISC,0
#define LOST_SIGNAL		MISC,1
#define	TEMPA			MISC,2
#define	TEMPB			MISC,3
#define	TEMPTOGGLE		MISC,4
#define PARITY_ERROR	MISC,5
#define	LOW_BATTERY		MISC,6

;PFM_FLAGS
#define	AILTOGGLE	PFM_FLAGS,0
#define	ELETOGGLE	PFM_FLAGS,1
#define	ESCTOGGLE	PFM_FLAGS,2
#define	RUDTOGGLE	PFM_FLAGS,3

;pins in PFM_PIN_BUF, arranged so we can copy to PORTC and line up all but rudder
#define	AIL_BUF_A		PFM_PIN_BUF,0
#define	AIL_BUF_B		PFM_PIN_BUF,1
#define	ESC_BUF_A		PFM_PIN_BUF,2
;
#define	ELE_BUF_A		PFM_PIN_BUF,3
#define	ELE_BUF_B		PFM_PIN_BUF,4
;#define	RUD_BUF_A		PFM_PIN_BUF,4
;#define	RUD_BUF_B		PFM_PIN_BUF,5
#define LBD_LED			d'5'		;Low Battery LED (actually on PORTC)

#define	AIL_PIN_A		PORTC,0
#define	AIL_PIN_B		PORTC,1
#define	ESC_PIN_A		PORTC,2
#define	ELE_PIN_A		PORTC,3
#define	ELE_PIN_B		PORTC,4
;#define	RUD_PIN_A		PORTC,4
;#define	RUD_PIN_B		PORTC,5


;masks to fetch just specific pins
#define AIL_PIN_MASK	b'00000011'
#define	ESC_PIN_MASK	b'00000100'
#define ELE_PIN_MASK	b'00011000'
#define	RUD_PIN_MASK	b'11000000'

;pin connections to IA 4320 RX chip
#define		IA_SDI		PORTA,0
#define		IA_SCK		PORTA,1
#define		IA_nSEL		PORTA,2
#define		IA_nIRQ		PORTA,3
#define		IA_SDO		PORTA,4


;IA4320 settings
;these require checking the datasheet, have fun
#define		IA_CONFIG		b'1001110000111001'			;915 Mhz band, low battery detect on, no wake-up timer or osc in sleep
														;XTL capacitance, 200khz baseband bw, disable clockout
#define		IA_FREQ			b'1010011111010000'			;915.00 Mhz
#define		IA_RX			b'1100000001000000'		 	;VDI=DQD, RX off
#define		IA_LBD			b'1100001000001000'			;set for 3.0V low battery detect, 1Mhz clkout
#define		IA_AFC			b'1100011001000111'			;enable AFC, one-shot at startup, unlimited offset, fine mode
#define		IA_DATAFILT		b'1100010001101100'			;fast clock lock, Digital filtering, DQD=4
#define		IA_DATARATE		b'1100100000100011'			;set expected bit rate as ~9600 (9579bps)
#define		IA_FIFO			b'1100111010000101'			;FFIT @ 8bits, start fill on sync word, disable  fill, enable fifo

#define		IA_FIFO_ENABLE	b'00000010'					;used to enable FIFO later (disabled for testing)
#define		IA_RX_ENABLE	b'00000001'					;used to enable RX later


;Define memory register aliases
 CBLOCK 0x020
	W_TEMP									;context save regs for ISR
	STATUS_TEMP

	;Values received from TX, these *must* adjacent in ascending order (use FSR/INDF later)
	INPUT_WORD0
	INPUT_WORD1
	INPUT_WORD2
	INPUT_WORD3
	INPUT_WORD_COUNT

	;buffers for data once we've had a successful read, to minimize the chance of "clobbering"
	AIL_INPUT
	ELE_INPUT
	ESC_INPUT
	RUD_INPUT

	;values for PFM
	AIL_PFM_VAL								;PFM set values
	ELE_PFM_VAL
	ESC_PFM_VAL
	RUD_PFM_VAL
	AIL_PFM_CNTR							;PFM counters
	ELE_PFM_CNTR
	ESC_PFM_CNTR
	RUD_PFM_CNTR
	PFM_FLAGS
	PFM_PIN_BUF								;buffer for desired pin settings

	SIGNAL_USE_COUNTER						;counters to track how often we re-use a signal
	SIGNAL_USE_COUNTER2
	TEMP
	PARITY									;used to check parity

	;SPI regs
	IA_DATAH
	IA_DATAL
	SPI_BIT_COUNTER
	SPI_TEMP

	MISC
 ENDC


;define Macros
;these are short enough to be more efficent as macros than subroutines
goBank1 MACRO
  bsf STATUS,RP0
  ERRORLEVEL -302							;acknowledge that we're in bank1, stop bank error warnings
 ENDM

goBank0 MACRO
  bcf STATUS,RP0
  ERRORLEVEL +302							;clear bank1 ack so we see banking error warnings again
 ENDM



;PROGRAM CODE STARTS HERE
 ORG 0  
  goto SETUP

;-------------------------------------
;interrupt service routine
 ORG 4
  ;context save
  MOVWF W_TEMP ;copy W to temp register, could be in either bank 
  SWAPF STATUS,W ;swap status to be saved into W 
  BCF STATUS,RP0 ;change to bank 0 regardless of current bank 
  MOVWF STATUS_TEMP ;save status to bank 0 register 

	;test if this was an IOC Interrupt
  btfsc INTCON,RAIF
  call HANDLEINTERRUPT

	;test if this was a TMR1 interrupt
  btfsc PIR1,TMR1IF				
  call HANDLETMR1IF				

  ;context restore
  SWAPF STATUS_TEMP,W;swap STATUS_TEMP register into W, sets bank to original state 
  MOVWF STATUS ;move W into STATUS register 
  SWAPF W_TEMP,F ;swap W_TEMP 
  SWAPF W_TEMP,W ;swap W_TEMP into W

	;return and re-enable interrupts
  retfie
;-------------------------------------


;-------------------------------------
ENABLEIOC
	;enable IOC interrupts.  Try not to disturb other interrupt settings
	;we hope everything else was set/cleared to sane states.
  goBank1
  movlw b'00001000'							;IOC enabled for GP3 (nIRQ)
  movwf IOC
  goBank0
  bcf INTCON,RAIF							;clear flag
  bsf INTCON,RAIE							;enable IOC
  bsf INTCON,GIE							;enable interrupts
  return
;-------------------------------------


;-------------------------------------
HANDLETMR1IF
  bcf PIR1,TMR1IF							;clear TMR1 interrupt flag
  goBank1
  bcf PIE1,TMR1IE							;disable TMR1 interrupt
  goBank0
	;If Timer1 expired, we must have lost the datastream, perhaps due to loss of signal in the RX chip, or bit shift
  call RESETDATASTREAM
  return
;-------------------------------------


;-------------------------------------
HANDLEINTERRUPT
	;handle an Interrupt from the IA4320 RX chip
  movfw PORTA								;have to read PORTA in order to reset RAIF
  bcf INTCON,RAIF							;clear the IOC flag
  btfsc IA_nIRQ								;test nIRQ for direction
  return									;only care about IOC 1->0 (interrupt pulse, active low)

	;it was an interrupt (0->1 on nIRQ)
  clrf IA_DATAH								;clear regs and carry bit for below
  clrf IA_DATAL
	;found an interrupt, read out the status bits and figure out what the RX chip is notifying us about
  bcf IA_SCK
  bcf IA_SDI								;read command is to send a CLK with SDI low (all other commands start high)
  bcf IA_nSEL								;nSEL set low to open SPI conversation
  nop										;min wait is 10 ns, 1 uS should do dandy ;)
	;Get the first 8 bits, which include the four interrupt bits (7-FFIT, 6-FFOV, 5-WKUP, 4-LBD)
  call get_8SPI_BITs
  movwf IA_DATAH
  btfsc IA_DATAH,6							;check FFOV
  goto RESETDATASTREAM						;FIFO overflow, Houston we have a problem...
  btfss IA_DATAH,7							;check if FFIT is set
  goto FINISH_SPI								;FFIT not set, no reason to keep reading
  call get_8SPI_BITs						;FFIT set, we should have 8 bits of data (after reading 8 more status bits)
  movwf IA_DATAL							;8 more status bits
	;Now the juicy bit, real data from the TX
  call get_8SPI_BITs
  movwf SPI_TEMP							;8 bits of data, w00t
  ;movfw SPI_TEMP	;redundant
  sublw HEADER								;test if the eight bits of data matches the header
  btfss STATUS,Z
  goto SAVE_8BITS							;not a match to header, regular data, go save
	;we received the header.  Reset things, as we appear to be getting a new frame [we may not have completed the last frame]
  clrf INPUT_WORD_COUNT
  incf INPUT_WORD_COUNT,f
  bcf T1CON,TMR1ON
  movlw d'256'-( (d'2'*BITS_PER_FRAME*XMIT_uSPB) % d'256')			;TMR1, with room for *twice* the datastream remainder
  movwf TMR1L														;if we expire the timer, something went horribly wrong
  movlw d'256'-( (d'2'*BITS_PER_FRAME*XMIT_uSPB) / d'256')
  movwf TMR1H
  bcf PIR1,TMR1IF							;clear TMR1 interrupt flag just in case
  bsf T1CON,TMR1ON
  bsf INTCON,PEIE							;enable Peripherial interrupts
  goBank1
  bsf PIE1,TMR1IE							;enable TMR1 interrupt
  goBank0
  goto FINISH_SPI							;we don't actually save the header...


SAVE_8BITS
  movfw INPUT_WORD_COUNT					;test if we the word counter == 0
  xorlw d'0'								;it shouldn't, unless we're receiving a non-header at the start of a frame
  btfsc STATUS,Z							;which we'd consider an error
  goto RESETDATASTREAM
  movlw INPUT_WORD0							;load address of word0
  movwf FSR
  decf INPUT_WORD_COUNT,w					;shift location according to the word count (-1 to zero-base it)
  addwf FSR,f
  movfw SPI_TEMP
  movwf INDF								;save using indirect addressing
  incf INPUT_WORD_COUNT,f					;increment to point to next data word storage location

FINISH_SPI
  nop
  bsf IA_nSEL	;finish SPI conversation
	;test LBD
  btfsc IA_DATAH,4
  bsf LOW_BATTERY							;turn on the Low battery LED

  movfw INPUT_WORD_COUNT
  sublw d'5'								;test if we've seen *exactly* the header + four words of data
  btfsc STATUS,Z							;this means extra data is ignored, could be used for more channels or optional
  goto FINISHED_FRAME						;settings/configuration data
  return


FINISHED_FRAME
	;we received a full four words of data, save it out to more permanent registers
  call RESETDATASTREAM						;reset the FIFO, TMR1 and counters
  movfw INPUT_WORD0
  movwf AIL_INPUT
  movfw INPUT_WORD1
  movwf ELE_INPUT
  movfw INPUT_WORD2
  movwf ESC_INPUT
  movfw INPUT_WORD3
  movwf RUD_INPUT
  bsf NEW_INPUTS
  return


RESETDATASTREAM
	;reset the datastream status, clear the FIFO, and it will look for the synchron pattern again
  bcf T1CON,TMR1ON							;disable TMR1, since we're not timing a stream anymore
  clrf INPUT_WORD_COUNT
  bcf NEW_INPUTS
	;make sure any current SPI conversation is over
  bsf IA_nSEL
	;Disable/Clear FIFO
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
	;Re-enable FIFO
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO) | IA_FIFO_ENABLE
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  return

;END HANDLEIOC
;-------------------------------------


;-------------------------------------
get_8SPI_BITs
	;fetch 8 bits
  movlw d'8'								;8 bits
  movwf SPI_BIT_COUNTER
  clrf SPI_TEMP
get_SPI_BIT
  rlf SPI_TEMP,f
  goto $+1
  bsf IA_SCK	;SCK rising edge
  goto $+1
  bcf SPI_TEMP,0							;pull in single bit off SDO of IA4320
  btfsc IA_SDO		
  bsf SPI_TEMP,0
  nop
  bcf IA_SCK	;SCK falling edge
  decfsz SPI_BIT_COUNTER,f
  goto get_SPI_BIT
  movfw SPI_TEMP
  goto $+1
  return
;-------------------------------------



;-------------------------------------
;Sub send_SPI
send_SPI
	;Sub to send specified # (in W) of bits via SPI (up to 16) from registers IA_DATAL & IA_DATAH
	;Starts with MSB of IA_DATAH, so bits should be left justified to IA_DATAH
  movwf SPI_BIT_COUNTER
  bcf IA_SCK
  bcf IA_SDI
  bcf IA_nSEL	;nSEL set low to open SPI conversation
  goto $+1
send_SPI_BIT
  bcf IA_SDI	;set value on SD output (SDI on IA chip)
  btfsc IA_DATAH,7
  bsf IA_SDI
  goto $+1
  bsf IA_SCK	;SCK rising edge
  rlf IA_DATAL,f
  rlf IA_DATAH,f
  bcf IA_SCK	;SCK falling edge
  decfsz SPI_BIT_COUNTER,f
  goto send_SPI_BIT
  goto $+1
  bsf IA_nSEL	;finish SPI conversation
  bcf IA_SDI
  return
;END send_SPI
;-------------------------------------


;-------------------------------------
WAIT256_W
	;this will wait 256 uS x W
  bcf INTCON,T0IF
  clrf TMR0									;TMR0 will overflow after W increments
  subwf TMR0,f
  goBank1
  movlw b'00000111'							;TMR0 config:prescaled 1:256
  movwf OPTION_REG
  goBank0
  btfss INTCON,T0IF	
  goto $-1
  bcf INTCON,T0IF
  return
;-------------------------------------


;-------------------------------------
SETUP
  clrf PORTA 								;clear outputs
  clrf PORTC								;clear outputs


  goBank1
  call 3FFh									
  movwf OSCCAL 								;config OSCCAL
  movlw b'00001000'							;TMR0 config: no prescaler, internal clocking (Fosc/4 = 1 cycle)
  movwf OPTION_REG
  clrf PIE1
#IFDEF __12F675
  clrf ANSEL								;turn off the 12F675's A/D unit
#ENDIF ;_12F675
  movlw b'00011000'							;set RA0,1,2,5 as outputs, 3&4 as inputs
  movwf TRISA
#IFDEF __16F630								;config 16F630's PORTC as outputs
  movlw b'00000000'							
  movwf TRISC
#ENDIF ;_16F630
  goBank0
  movlw b'00000001'							;TMR1 config: no prescaler, internal clocking (Fosc/4 = 1 cycle)
  movwf T1CON
  bcf T1CON,TMR1ON							;disable TMR1
  movlw 07h									;Turn off Comparator unit (111)
  movwf CMCON

  clrf MISC

  clrf INTCON								;clear interrupt settings
  bcf NEW_INPUTS							;clear flags
  bsf LOST_SIGNAL							;start in a "lost signal" fashion

;done PIC setup


;do IA4320 setup
  bsf IA_nSEL		;it's logic low, so set it high

;sleep for a bit so the RX chip can stabilize after it's POR
  movlw d'1000000'/d'65536'					;one second, rounded down
  movwf TEMP
OUTER_LOOP
  movlw d'255'
  call WAIT256_W
  decfsz TEMP,f
  goto OUTER_LOOP

	;Receiver Setting *shut off RX*
  movlw HIGH(IA_RX)
  movwf IA_DATAH
  movlw LOW(IA_RX)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Configuration Setting
  movlw HIGH(IA_CONFIG)
  movwf IA_DATAH
  movlw LOW(IA_CONFIG)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI 
  movlw d'2'
  call WAIT256_W
	;Frequency Setting
  movlw HIGH(IA_FREQ)
  movwf IA_DATAH
  movlw LOW(IA_FREQ)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Wake up timer left at default
	;Low duty cycle left at default
	;Low battery threshold & clockout
  movlw HIGH(IA_LBD)
  movwf IA_DATAH
  movlw LOW(IA_LBD)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Data Filter
  movlw HIGH(IA_DATAFILT)
  movwf IA_DATAH
  movlw LOW(IA_DATAFILT)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Data Rate setting
  movlw HIGH(IA_DATARATE)
  movwf IA_DATAH
  movlw LOW(IA_DATARATE)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;AFC command
  movlw HIGH(IA_AFC)
  movwf IA_DATAH
  movlw LOW(IA_AFC)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Receiver Setting & enable RX
  movlw HIGH(IA_RX)
  movwf IA_DATAH
  movlw LOW(IA_RX) | IA_RX_ENABLE
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Output/FIFO setting
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W

	;Wait for Oscillator & PLL to stabilize
  movlw d'10'
  call WAIT256_W

	;Enable FIFO now
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO) | IA_FIFO_ENABLE
  movwf IA_DATAL
  movlw d'16'
  call send_SPI


  goBank1
  movlw b'00001000'							;TMR0 config: no prescaler, internal clocking (Fosc/4 = 1 cycle)
  movwf OPTION_REG
  goBank0
  movlw PFM_FREQ
  movwf TMR0

	;use interrupts
  ;call ENABLEIOC

;-------------------------------------
MAIN
	;main body of the code
	;we loop in here forever

  ;poll interrupt pin (active low)
  btfss IA_nIRQ
  ;btfss IA_SDO ;SDO pin goes low on FFIT, doesn't give is IT for other interrupts tho...need housekeeping routine
  call HANDLEINTERRUPT

;for testing PFM
; MESSG "Test data deployed, this code won't actually work!"
;  movlw b'10101000'	
;  movwf AIL_INPUT
;  movlw b'01110000'	
;  movwf ELE_INPUT
;  movlw b'00011100'	
;  movwf ESC_INPUT
;  movlw b'10101000'	
;  movwf RUD_INPUT


  btfsc NEW_INPUTS
  call PARSE_INPUTS

	;we don't use the interrupt on TMR0 overflow, but we can still test on the flag
  btfss INTCON,T0IF
  goto MAIN

	;TMR0 expired, set counter again and do some PFM if we have recent signals
  movlw PFM_FREQ
  movwf TMR0
  bcf INTCON,T0IF
  btfss LOST_SIGNAL
  call DOPFM								;do PFM on the outputs

  goto MAIN
;END MAIN
;-------------------------------------




;-------------------------------------
PARSE_INPUTS
  bcf NEW_INPUTS

  movfw AIL_INPUT
  call TEST_PARITY						;test parity, if there's a parity error, ignore the data
  btfsc PARITY_ERROR					;we'll end up re-using the value from last time
  goto PARSE_ELE						;We hope we get at least *some* usable info from each frame
  bcf AILTOGGLE
  bcf AIL_BUF_A
  bcf AIL_BUF_B
  call PARSEINPUT
  movwf AIL_PFM_VAL
  btfsc TEMPA
  bsf AIL_BUF_A
  btfsc TEMPB
  bsf AIL_BUF_B
  btfsc TEMPTOGGLE
  bsf AILTOGGLE

PARSE_ELE
  movfw ELE_INPUT
  call TEST_PARITY
  btfsc PARITY_ERROR
  goto PARSE_ESC
  bcf ELETOGGLE
  bcf ELE_BUF_A
  bcf ELE_BUF_B
  call PARSEINPUT
  movwf ELE_PFM_VAL
  btfsc TEMPA
  bsf ELE_BUF_A
  btfsc TEMPB
  bsf ELE_BUF_B
  btfsc TEMPTOGGLE
  bsf ELETOGGLE

	;esc is a bit different
PARSE_ESC
  bcf ESCTOGGLE
  bcf ESC_BUF_A
  movfw ESC_INPUT
  call TEST_PARITY
  btfsc PARITY_ERROR
  goto PARSE_RUD
  bsf ESC_BUF_A
  andlw b'01111111'				;clear out the parity bit
  sublw d'127'					;pivot value around 127 to get PFM frequency
  movwf ESC_PFM_VAL
  sublw d'5'					;test if it's maxed
  btfsc STATUS,C
  goto PARSE_RUD				;esc maxed, leave it on with toggling disabled
  movfw ESC_PFM_VAL
  sublw d'127'-d'20'			;test if it's mined  (rather generous due to my hitec test radio)
  btfss STATUS,C
  goto ESC_MIN
  bsf ESCTOGGLE					;esc in normal pfm range, enable pfm toggling
  goto PARSE_RUD
ESC_MIN
  bcf ESC_BUF_A					;esc at min, turn off the pin with toggling disabled


PARSE_RUD
#ifdef __16F630_FIXMESOMEDAY
  movfw RUD_INPUT
  call TEST_PARITY
  btfsc PARITY_ERROR
  goto FLAG_GOOD_INPUT
  bcf RUDTOGGLE
  bcf RUD_BUF_A
  bcf RUD_BUF_B
  call PARSEINPUT
  movwf RUD_PFM_VAL
  btfsc TEMPA
  bsf RUD_BUF_A
  btfsc TEMPB
  bsf RUD_BUF_B
  btfsc TEMPTOGGLE
  bsf RUDTOGGLE

#endif

	;prime PFM counters
  movlw PFM_ON_TIME
  movwf AIL_PFM_CNTR
  movwf ELE_PFM_CNTR
  movwf ESC_PFM_CNTR
  movwf RUD_PFM_CNTR


	;set Low battery LED on if needed
  bcf PFM_PIN_BUF,LBD_LED
  btfsc LOW_BATTERY
  bsf PFM_PIN_BUF,LBD_LED

	;now move new pin values to PORTC
  movfw PFM_PIN_BUF
  movwf PORTC


FLAG_GOOD_INPUT
	;received a set of good inputs, reset the lost-signal counters
  bcf LOST_SIGNAL
  movlw SIGNAL_USE_THRESHOLD
  movwf SIGNAL_USE_COUNTER
  movlw d'255'
  movwf SIGNAL_USE_COUNTER2

  return
;END PARSE_INPUTS
;-------------------------------------



;-------------------------------------
TEST_PARITY
  ;Calculate even parity for bits 0-6 of W
  ;and compares against existing parity flag (bit 7 of W)
  ;sets PARITY_ERROR if the result is an error, clears otherwise
  bcf PARITY_ERROR
  movwf TEMP
  clrf PARITY
  btfsc TEMP,0
  incf PARITY,f
  btfsc TEMP,1
  incf PARITY,f
  btfsc TEMP,2
  incf PARITY,f
  btfsc TEMP,3
  incf PARITY,f
  btfsc TEMP,4
  incf PARITY,f
  btfsc TEMP,5
  incf PARITY,f
  btfsc TEMP,6
  incf PARITY,f


	;test the existing parity flag
  btfss TEMP,7
  goto ODD
	;else we have an even # of 1's, so PARITY,0 should be clear
  btfsc PARITY,0
  bsf PARITY_ERROR
  return
ODD
	;else we have an odd # of 1's, so PARITY,0 should be set
  btfss PARITY,0
  bsf PARITY_ERROR
  return

;END TEST_PARITY
;-------------------------------------


;-------------------------------------
PARSEINPUT
;convert value from 7-bit [0-127] directional value to [0-63] strength value & direction bit
;including setting pin direction and toggling flags
;expect inputs in W, writes modified value out to W
; and sets/clears flags TEMPPOS, TEMPA, TEMPB
  bcf TEMPA
  bcf TEMPB
  bsf TEMPTOGGLE
  movwf TEMP
  bcf TEMP,7						;clear parity flag, it's already been checked
  btfss TEMP,6						;test if it's >64, ie. over half
  goto CH_NEG						;under half, we can use the raw value
	;channel is positive, subtract from 127
  movfw TEMP
  bsf TEMPA
  sublw d'127'
  movwf TEMP
  goto TEST_MARGINS							
CH_NEG 					
  bsf TEMPB
TEST_MARGINS										
  ;now we have a strength only value we can use for PFM
  ;where max is ~0, and min is ~63 (and a flag indicates actual direction)
  ;now test against our max/min bounds (technically, min will be around center)
  sublw d'5'
  btfsc STATUS,C
  bcf TEMPTOGGLE					;max value, turn it on, full time (no toggling)
  movlw d'63'-d'3'
  subwf TEMP,w
  btfss STATUS,C
  goto PARSEEND
  bcf TEMPA							;min value = center-ish, turn it off
  bcf TEMPB
  bcf TEMPTOGGLE
PARSEEND
  movfw TEMP						;move the calculated value back to W
  return
;END PARSEINPUT
;-------------------------------------




;-------------------------------------
DOPFM
  decfsz SIGNAL_USE_COUNTER2,f
  goto NORMALPFM
  decfsz SIGNAL_USE_COUNTER,f
  goto NORMALPFM
	;if we get here, the signal has been over-used without being refreshed
	;go into "lost-signal" mode by clearing the outputs
  bsf LOST_SIGNAL
  clrf PORTC
  return

NORMALPFM
	;strategy with PFM is to test if we need to toggle the channel (min/max stay full off/on)
	;if so, decrement and check the counter, toggling when the counter expires
  btfss AILTOGGLE
  goto TEST_ELE
  decfsz AIL_PFM_CNTR,f
  goto TEST_ELE										;counter still hasn't expired, move on
  btfsc AIL_PIN_A
  goto TOGGLEAILOFF
  btfsc AIL_PIN_B
  goto TOGGLEAILOFF
	;have to turn correct ail pin on
  movfw PFM_PIN_BUF									;fetch desired pin settings
  andlw AIL_PIN_MASK								;mask out correct pins
  iorwf PORTC,f										;and ior to PORTA
  movlw PFM_ON_TIME									;set cntr for desired on time
  movwf AIL_PFM_CNTR
  goto TEST_ELE
TOGGLEAILOFF
  bcf AIL_PIN_A										;clear pins and set counter
  bcf AIL_PIN_B
  movfw AIL_PFM_VAL
  movwf AIL_PFM_CNTR


TEST_ELE
  btfss ELETOGGLE
  goto TEST_ESC
  decfsz ELE_PFM_CNTR,f
  goto TEST_ESC										;counter still hasn't expired, move on
  btfsc ELE_PIN_A
  goto TOGGLEELEOFF
  btfsc ELE_PIN_B
  goto TOGGLEELEOFF
	;have to turn correct ail pin on
  movfw PFM_PIN_BUF									;fetch desired pin settings
  andlw ELE_PIN_MASK								;mask out correct pins
  iorwf PORTC,f										;and ior to PORTA
  movlw PFM_ON_TIME									;set cntr for desired on time
  movwf ELE_PFM_CNTR
  goto TEST_ESC
TOGGLEELEOFF
  bcf ELE_PIN_A										;clear pins and set counter
  bcf ELE_PIN_B
  movfw ELE_PFM_VAL
  movwf ELE_PFM_CNTR

TEST_ESC
  btfss ESCTOGGLE
  goto TEST_RUD
  decfsz ESC_PFM_CNTR,f
  goto TEST_RUD										;counter still hasn't expired, move on
  btfsc ESC_PIN_A
  goto TOGGLEESCOFF
	;have to turn correct ail pin on
  movfw PFM_PIN_BUF									;fetch desired pin settings
  andlw ESC_PIN_MASK								;mask out correct pins
  iorwf PORTC,f										;and ior to PORTA
  movlw PFM_ON_TIME									;set cntr for desired on time
  movwf ESC_PFM_CNTR
  goto TEST_RUD
TOGGLEESCOFF
  bcf ESC_PIN_A										;clear pins and set counter
  movfw ESC_PFM_VAL
  movwf ESC_PFM_CNTR

TEST_RUD
#ifdef __16F630_FIXME_SOMEDAY
  btfss RUDTOGGLE
  return
  decfsz RUD_PFM_CNTR,f
  return										;counter still hasn't expired, move on
  btfsc RUD_PIN_A
  goto TOGGLERUDOFF
  btfsc RUD_PIN_B
  goto TOGGLERUDOFF
	;have to turn correct ail pin on
  movfw PFM_PIN_BUF									;fetch desired pin settings
  andlw RUD_PIN_MASK								;mask out correct pins
  iorwf PORTC,f										;and ior to PORTA
  movlw PFM_ON_TIME									;set cntr for desired on time
  movwf RUD_PFM_CNTR
  return
TOGGLERUDOFF
  bcf RUD_PIN_A										;clear pins and set counter
  bcf RUD_PIN_B
  movfw RUD_PFM_VAL
  movwf RUD_PFM_CNTR
#endif
  return
;END DOPFM
;-------------------------------------

 END

--

 END

w PFM_ON_TIME									;set cntr for desired on time
  movwf RUD_PFM_CNTR
  return
TOGGLERUDOFF
  bcf RUD_PIN_A										;clear pins and set counter
  bcf RUD_PIN_B
  movfw RUD_PFM_VAL
  movwf RUD_PFM_CNTR
#endif
  return
;END DOPFM
;-------------------------------------

 END

--

 END

-----

 END

--

 END

