;DIY-UHF Digital signal Encoding TX
;Version 1.1
;David Lancaster
;All rights reserved

;History
;1.1 Moved to using 5Mhz clkout from IA4220 chip for more accurate bit timing



;CONFIGURABLE OPTIONS

;Transmitter type 
;uncomment the correct one, comment out the others
;#DEFINE JR
#DEFINE FUTABA
;#DEFINE SANWA

;Set Logic levels, comment for low-level logic
;#DEFINE LOGIC_HIGH


;END CONFIGURABLE OPTIONS


	
;some limits to the RC PPM signal, exceptions are treated as errors
;Note that the channel width does *not* include the pulse width
;these may need fine tuning for specific radios/brands, value in mS

#define		MAX_CHANNEL_WIDTH			d'2050'		;maximum width of a channel
#define		MIN_CHANNEL_WIDTH			d'950'		;minimum width of a channel
#define		MIN_PULSE_WIDTH				d'300'		;minimum width of a pulse
#define		MAX_PULSE_WIDTH				d'550'		;maximum width of a pulse







#IFDEF __12F629
	MESSG "Assembling for 12F629"
	LIST	P=PIC12F629
	INCLUDE "P12F629.INC"
	__CONFIG _MCLRE_OFF & _CP_OFF & _WDT_OFF & _EC_OSC & _CPD_OFF & _PWRTE_ON & _BODEN_ON
	;ERRORLEVEL -302					;turn off register not in bank0 warnings
#ENDIF

#IFDEF __12F675
	MESSG "Assembling for 12F675"
	LIST	P=PIC12F675
	INCLUDE "P12F675.INC"
	__CONFIG _MCLRE_OFF & _CP_OFF & _WDT_OFF & _EC_OSC & _CPD_OFF & _PWRTE_ON & _BODEN_ON
	;ERRORLEVEL -302					;turn off register not in bank0 warnings
#ENDIF

#IFDEF __16F630
	MESSG "Assembling for 16F630"
	LIST	P=PIC16F630
	INCLUDE "P16F630.INC"
	__CONFIG _MCLRE_OFF & _CP_OFF & _WDT_OFF & _EC_OSC & _CPD_OFF & _PWRTE_ON & _BODEN
    #DEFINE GPIO PORTA  	;Porta is equiv to GPIO
	#DEFINE TRISIO TRISA 	;Trisa is equiv to TRISIO
	;ERRORLEVEL -302					;turn off register not in bank0 warnings
#ENDIF


;DEFINE CONSTANTS
;----------------

#INCLUDE ../data_format.inc

#define		CLOCKSPEED				d'5'
#define		CLKSKEW					CLOCKSPEED/d'4'		;Fosc in Mhz/4 (1op per 4 osc)


#define		c3MS						d'3000'
#define		c17MS						d'17000'
#define		MAX_CHANNEL_WIDTH_TICKS		MAX_CHANNEL_WIDTH*CLKSKEW
#define		MIN_CHANNEL_WIDTH_TICKS		MIN_CHANNEL_WIDTH*CLKSKEW
#define		MIN_PULSE_WIDTH_TICKS		MIN_PULSE_WIDTH*CLKSKEW
#define		MAX_PULSE_WIDTH_TICKS		MAX_PULSE_WIDTH*CLKSKEW


#define		OUTPIN					GPIO,0
#define		INPUT_PIN				GPIO,3

#define		NUMERRORS				XMIT_BPS/d'140'	;about 1/2 second, give or take

#define		INPUT_ERROR_FLAG		MISC, 7	;Flag to indicate input data is invalid
#define		SEND_DATA_FLAG			MISC, 6	;Flag to indicate that we can send data



;some register reuse we can get away with early on
#define		IA_DATAL	TEMP
#define		IA_DATAH	PARITY
#define		IA_BTS		ESC_INPUT	;bits to send

;pin connections to IA 422X TX chip
#define		IA_SDI		GPIO,1
#define		IA_SCK		GPIO,2
#define		IA_nSEL		GPIO,5
#define		IA_FSK		GPIO,0	;same as OUTPIN
#define		IA_nIRQ		GPIO,4	;not used

;some Integrated settings for IA432X RX
#define 	PREAMBLE	0xAA
#define		SYNCHRON	0x2DD4
;the RX expects 16-bits of preamble to warm up, followed by the 16-bit syncron to start FIFO buffering, followed by data
#define		IA_SYNC_LEN	d'32'

;IA4220 settings
#define		IA_PWRMGT_OFF	b'1100000000000000'		;turn off units
#define		IA_PWRMGT_ON	b'1100000000111000'		;Manually control: Turn on crystal osc, synthesizer & power amp  & clock output, disable low-battery
#define		IA_CONFIG		b'1001111000110010'		;915 Mhz band, 5Mhz clkout	;Set Freq deviation to 90 khz
#define		IA_FREQ			0xA7D0					;915.00 Mhz
#define		IA_PWR			b'10110000'				;FSK signalling, high power



BIGWAIT MACRO
	;this is 65536 ticks, times uS_perTick
  clrf TMR0									;TMR0 will overflow after 256 increments
  goBank1							;Set config options in Bank1
  bsf OPTION_REG,PS2						;Set TMR0 for 1:256 prescaler (111)
  bsf OPTION_REG,PS1
  bsf OPTION_REG,PS0
  goBank0
  btfss INTCON,T0IF	
  goto $-1
  bcf INTCON,T0IF
 ENDM

goBank1 MACRO
  bsf STATUS,RP0
  ERRORLEVEL -302
 ENDM

goBank0 MACRO
  bcf STATUS,RP0
  ERRORLEVEL +302
 ENDM

;VARIABLES
;---------

#define CBVAL 0x020		;12F6XX/16F6XX regs start at 20h

 CBLOCK CBVAL						;Start a Variable block, starting at CBVAL set above
MISC								;misc flag holder

;Holds values received from TX
ESC_INPUT							;ESC buffer
AIL_INPUT							;AILERON buffer
ELE_INPUT							;Elevator buffer
RUD_INPUT							;Rudder buffer

CURRENTCHANNEL						;holds currentchannel-1 when reading input
CHANNELVALUE						;hold value of current channel when reading input

;Holds values to send in SENDDATA
ESC_VAL								;ESC value
AIL_VAL								;AILERON value
ELE_VAL								;ELEVATOR value
RUD_VAL								;RUDDER value
;used in SENDDATA
CURRENT								;registers to hold data as we send it
PRE2
SYNCL
SYNCH
HEAD
TOTALBITCNTR						;number of bits to send
ERROR_COUNTER						;used to count transmit cycles without valid input
;used to calculate parity
PARITY								;counter to track parity
TEMP								;temp register

PWRVAL

 ENDC

 ERRORLEVEL -207
;organize channels so that the values go where we expect them to
#IFDEF JR
;Config for JR, ESC, AIL, ELE, RUD
	CH1_INPUT	EQU ESC_INPUT
	CH2_INPUT	EQU AIL_INPUT
	CH3_INPUT	EQU ELE_INPUT
	CH4_INPUT	EQU RUD_INPUT
#ENDIF ;JR

#IFDEF FUTABA
;Config for Futaba/Hitec, AIL, ELE, ESC, RUD
	CH1_INPUT	EQU AIL_INPUT
	CH2_INPUT	EQU ELE_INPUT
	CH3_INPUT	EQU ESC_INPUT
	CH4_INPUT	EQU RUD_INPUT
#ENDIF ;FUTABA

#IFDEF SANWA
;Config for SANWA, AIL, ESC, ELE, RUD
	CH1_INPUT	EQU AIL_INPUT
	CH2_INPUT	EQU ESC_INPUT
	CH3_INPUT	EQU ELE_INPUT
	CH4_INPUT	EQU RUD_INPUT
#ENDIF ;SANWA


#IFDEF LOGIC_HIGH
	;pulse defined as a digital 1
	btpulsesc macro
			btfsc INPUT_PIN	
	endm
	btpulsess macro
			btfss INPUT_PIN	
	endm					
#ELSE
	;pulse defined as a digital 0
	btpulsesc macro
			btfss INPUT_PIN	
	endm
	btpulsess macro
			btfsc INPUT_PIN	
	endm							
#ENDIF
 ERRORLEVEL +207



;PROGRAM CODE STARTS HERE
 ORG 0  
  Goto SETUP


;-------------------------------------
SENDDATA
;START SUBROUTINE SENDDATA

	;test how valid the data we've got is
  btfss INPUT_ERROR_FLAG
  goto NORMALSEND
	;if we've gotten here, we're having problems with the input data
  decfsz ERROR_COUNTER,f
  goto NORMALSEND
	;if we've gotten here, we haven't had good input for quite some time, stop transmitting the bitstream
	;so that the RX can drop into failsafe (the carrier signal is still present)
  bcf SEND_DATA_FLAG
  return

NORMALSEND
	;send the data, starting with MSB
	;we send the Integrated header, our header, Aileron, Elevator, Throttle, and Rudder in that order
  movlw BITS_PER_FRAME+IA_SYNC_LEN+1
  movwf TOTALBITCNTR
  movlw PREAMBLE
  movwf CURRENT
  movwf PRE2
  movlw LOW(SYNCHRON)
  movwf SYNCL
  movlw HIGH(SYNCHRON)
  movwf SYNCH
  movlw HEADER
  movwf HEAD

  call WAITBIT				;this call is to setup TMR0 correctly, does involve a "bit" of a wait tho (ha...ha...ha...)

SENDABIT 
  btfsc CURRENT,7
  goto $+3
  bcf OUTPIN
  goto $+2
  bsf OUTPIN

  call WAITBIT

	;line arrays up beside each other, and do a huge global shift of one bit
  bcf STATUS,C
  rlf RUD_VAL,f
  rlf ESC_VAL,f
  rlf ELE_VAL,f
  rlf AIL_VAL,f
  rlf HEAD,f
  rlf SYNCL,f
  rlf SYNCH,f
  rlf PRE2,f
  rlf CURRENT,f

  decfsz TOTALBITCNTR,f
  goto SENDABIT

  bcf OUTPIN
  return

;END SENDDATA
;-------------------------------------



;-------------------------------------
WAITBIT
	;wait until TMR0 expires
  btfss INTCON,T0IF
  goto $-1

	;reset TMR0 & clear interrupt flag (we're not using interrupts, but the flag still works
  bcf INTCON,T0IF
  movlw d'256'-(XMIT_uSPB*CLKSKEW)+7
  movwf TMR0
  goBank1									;Set config options in Bank1
  movlw b'00001000'							;TMR0 config: prescaler assigned to WDT, internal clocking 
  movwf OPTION_REG
  goBank0

  return
;END WAIT
;-------------------------------------



;-------------------------------------
READ_INPUT
;START SUBROUTINE READ_INPUT
	;*****************************************
		;read input from RC TX, parsing out the values for the first four channels
		;uses some hard coded max/mins & some defines
	;*****************************************
	;Loop until we have no pulses for at least 3ms, using TMR0
  bcf INTCON,T0IF
  movlw d'256'-((c3MS*CLKSKEW)/d'16')				;calculate 3ms in TMR0 ticks, prescaled by 1:16
  movwf TMR0
  movlw b'00000011'
  goBank1							;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0 by 1:16
  goBank0
WAIT3ms
  btfsc INTCON,T0IF
  goto STARTREADING
  btpulsess
  goto WAIT3ms
  goto READ_INPUT		;hrm, technically, we can get stuck in here with no signal, but who cares

STARTREADING
	;Wait for a pulse in the next 17ms, using TMR0.  Error out after 17ms
  bcf INTCON,T0IF
  movlw d'256'-((c17MS*CLKSKEW)/d'128')			;calculate 17ms in TMR0 ticks, prescaled by 1:128
  movwf TMR0
  movlw b'00000110'
  goBank1							;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0 by 1:128
  goBank0
WAIT17ms
  btfsc INTCON,T0IF
  goto INPUT_ERROR
  btpulsess
  goto WAIT17ms

	;set channel counter to first channel (n-1)
  movlw d'0'
  movwf CURRENTCHANNEL

READPULSE
	;We count the channel on TMR0, treating > MAX_CHANNEL_WIDTH as an error
	;start our channel timer here
  bcf INTCON,T0IF
  movlw d'256'-((MAX_CHANNEL_WIDTH_TICKS+d'100')/d'16')
  movwf TMR0
  movlw b'00000011'
  goBank1									;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0 by 1:8
  goBank0
WAITONPULSE
  btfsc INTCON,T0IF							;if TMR0 overflowed, the pulse was *way* too long
  goto INPUT_ERROR
  btpulsesc
  goto WAITONPULSE
	;if we get here, the pulse ended, test for minimum length
  movlw (MIN_PULSE_WIDTH_TICKS)/d'16'
  subwf TMR0,W
  btfss STATUS,C
  goto INPUT_ERROR
	;keep tracking time until the beginning of the next pulse
WAITONCHAN
  btfsc INTCON,T0IF							;if TMR0 overflows, the chan was too long
  goto INPUT_ERROR
  btpulsess
  goto WAITONCHAN
	;now we have a value for the channel, quickly save it and start the timer for the next channel
  movfw TMR0
  movwf CHANNELVALUE
	;set TMR0 so it starts counting from *NOW* (which is actually a few us late)
  bcf INTCON,T0IF
  movlw d'256'-((MAX_CHANNEL_WIDTH_TICKS+d'100')/d'16') 
  movwf TMR0
  movlw b'00000011'
  goBank1									;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0 by 1:16
  goBank0

	;now that we have a little more time, restructure the values into something more sensical
  movlw d'256'-((MAX_CHANNEL_WIDTH_TICKS)/d'16') + (MIN_CHANNEL_WIDTH_TICKS/d'16') 	;this converts the number from TMR0_base + N, to just N
  subwf CHANNELVALUE,f
  btfss STATUS,C							;test if it went negative
  clrf CHANNELVALUE							;and clear if necessary						
	;now we should have just the interesting part of the channel length in ticks, with full low stick=0 and full high stick=k
  bcf STATUS,C
  rlf CHANNELVALUE,f						;multiply by 2

  movfw CHANNELVALUE
  sublw MAX_CHANNEL_WIDTH_TICKS/d'16' 		;now test if the remainder is over the maximum threshold	
  btfsc STATUS,C
  goto SAVECHAN
  movlw MAX_CHANNEL_WIDTH_TICKS/d'16'-1 	;and set it to max if necessary
  movwf CHANNELVALUE



SAVECHAN
	;now rescale the value from 0-127
  pageselw LOOKUPTABLE						;Lookuptable is in another page, and uses a computed goto, set PCLATH
  movfw CHANNELVALUE
  call LOOKUPTABLE
  movwf CHANNELVALUE
  clrf PCLATH
	;and save
  movfw CURRENTCHANNEL
  incf CURRENTCHANNEL,f
  addwf PCL,f
  goto CH1
  goto CH2
  goto CH3
  goto CH4

CH1
  movfw CHANNELVALUE
  movwf CH1_INPUT
  goto WAITONPULSE
CH2
  movfw CHANNELVALUE
  movwf CH2_INPUT
  goto WAITONPULSE
CH3
  movfw CHANNELVALUE
  movwf CH3_INPUT
  goto WAITONPULSE
CH4
  movfw CHANNELVALUE
  movwf CH4_INPUT

	;done reading channels, set some flags
  bcf INPUT_ERROR_FLAG			;clear error flag if it's set
  bsf SEND_DATA_FLAG			;assert that we want to be sending data these days
  movlw NUMERRORS				;and reset the error counter
  movwf ERROR_COUNTER
  return

INPUT_ERROR
	;Error reading input
	;Pin exceeded timeout or pulse width, etc
  bsf INPUT_ERROR_FLAG
  return

;-------------------------------------
;END READ_INPUT


;-------------------------------------
;Sub send_SPI
	;some register reuse, since send_SPI is called before we try reading input
#define	SPI_BIT_COUNTER ESC_VAL
send_SPI
	;Sub to send specified # of bits (in W) via SPI (up to 16) from registers IA_DATAL & IA_DATAH
	;Starts with MSB of IA_DATAH, so bits should be left justified to IA_DATAH
  movwf SPI_BIT_COUNTER
  bcf IA_SCK
  bcf IA_SDI
  bcf IA_nSEL	;nSEL set low to open SPI conversation
  goto $+1
send_SPI_BIT
  bcf IA_SDI	;set value on SD output (SDI on IA chip)
  btfsc IA_DATAH,7
  bsf IA_SDI
  goto $+1
  bsf IA_SCK	;SCK rising edge
  goto $+1
  rlf IA_DATAL,f
  rlf IA_DATAH,f
  bcf IA_SCK	;SCK falling edge
  decfsz SPI_BIT_COUNTER,f
  goto send_SPI_BIT
  goto $+1
  bsf IA_nSEL	;finish SPI conversation
  goto $+1
  bcf IA_SDI
  return
;END send_SPI
;-------------------------------------



;-------------------------------------

SETUP
	;NOTE: Since we're using the IA4220 for clock, note that it starts at 1Mhz, which we later change
  clrf GPIO 								;clear outputs
#IFDEF __16F630
  clrf PORTC								;clear outputs
#ENDIF ;_16F630


  goBank1									;Set config options in Bank1
  call 3FFh									
  movwf OSCCAL 								;config OSCCAL
  movlw b'00000000'							;TMR0 config at default
  movwf OPTION_REG
#IFDEF __12F675
  clrf ANSEL								;turn off the 12F675's A/D unit
#ENDIF ;_12F675
  movlw b'00001000'							;set GP3 to input, GP0-2,4,5 as outputs
  movwf TRISIO
#IFDEF __16F630								;config 16F630's PORTC as outputs
  movlw b'00000000'							
  movwf TRISC
#ENDIF ;_16F630
  goBank0									;done with Bank1 config options
  movlw 07h									;Turn off Comparator unit (111)
  movwf CMCON

  bcf SEND_DATA_FLAG						;make sure we don't suffer premature sendulation

;done PIC setup

;do IA422X setup
  bsf IA_nSEL		;it's logic low, so set it high

;sleep for a bit so the TX chip can stabilize after it's POR
  movlw d'250000'/d'65536'					;one second, rounded down
  movwf TEMP
OUTER_LOOP
  BIGWAIT	;this label is a lie, at this clock rate, it's actually four times as long
  decfsz TEMP,f
  goto OUTER_LOOP

;these require checking the IA422X datasheet, have fun
	;Power Management Setting *shut down tx first*
  movlw HIGH(IA_PWRMGT_OFF)
  movwf IA_DATAH
  movlw LOW(IA_PWRMGT_OFF)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  BIGWAIT
	;Configuration Setting
  movlw HIGH(IA_CONFIG)
  movwf IA_DATAH
  movlw LOW(IA_CONFIG)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI 
	;Frequency Setting 915
  movlw HIGH(IA_FREQ)
  movwf IA_DATAH
  movlw LOW(IA_FREQ)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
	;Power Setting
  movlw IA_PWR
  movwf IA_DATAH
  movlw d'8'
  call send_SPI
	;Power Management Setting
  movlw HIGH(IA_PWRMGT_ON)
  movwf IA_DATAH
  movlw LOW(IA_PWRMGT_ON)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
	;No changes to 	Bit Rate, Low Battery, Sleep, push-button, or wake-up timer config
	;Wait for Oscillator & PLL to stabilize
  nop
  BIGWAIT
  ;clrf PWRVAL

MAIN
	;main body of the code
	;we loop in here forever
  call READ_INPUT

  ;move latest values to output buffers
  movfw AIL_INPUT
  call LAZY_PARITY
  movwf AIL_VAL

  movfw ELE_INPUT
  call LAZY_PARITY
  movwf ELE_VAL

  movfw ESC_INPUT
  call LAZY_PARITY
  movwf ESC_VAL

  movfw RUD_INPUT
  call LAZY_PARITY
  movwf RUD_VAL

  btfsc SEND_DATA_FLAG
  call SENDDATA				;only send data if the flag is set

  goto MAIN
;END of main program, loop back
;-------------------------------------


LAZY_PARITY
;-------------------------------------
  ;Calculate even parity for bits 0-6 of W
  ;Parity stored in bit 7
  #define PARITY_POSITION	7
  movwf TEMP
  clrf PARITY
  btfsc TEMP,0
  incf PARITY,f
  btfsc TEMP,1
  incf PARITY,f
  btfsc TEMP,2
  incf PARITY,f
  btfsc TEMP,3
  incf PARITY,f
  btfsc TEMP,4
  incf PARITY,f
  btfsc TEMP,5
  incf PARITY,f
  btfsc TEMP,6
  incf PARITY,f


	;if # of 1's is even, PARITY,0 will be clear: set parity flag
	;if # of 1's is odd,  PARITY,0 will be set: clear parity flag
  bcf TEMP,PARITY_POSITION
  btfss PARITY,0
  bsf TEMP,PARITY_POSITION

  movfw TEMP
  return
;END LAZY_PARITY
;-------------------------------------



variable i
i=0
;Build a lookup table to convert ticks into a 0-127 range
;we expect the offset to be in W
LOOKUPTABLE
  addwf PCL,f
while  i<(MAX_CHANNEL_WIDTH_TICKS/d'16')
  retlw (i*d'4'/CLOCKSPEED)	;invert the tick skew
i+=1
endw
;some spare padding in case we jump a few past the lookup table
  retlw 0x7f
  retlw 0x7f
  retlw 0x7f
  retlw 0x7f
  retlw 0x7f

  END

tlw 0x7f
  retlw 0x7f
  retlw 0x7f
  retlw 0x7f

  END

